//###########################################################################
//
// C2000 Software SCI Example
// Simulating GPIO as SCI interface with XINT & CPU_timer
//
// RX - GPIO0     : SCI RX Pin
// TX - GPIO33    : SCI TX Pin
// Test - GPIO32  : Test pin, logic 1 indicates software SCI timer interrupt is triggered
//
//###########################################################################
// $TI Release: F2806x C/C++ Header Files and Peripheral Examples V151 $
// $Release Date: February  2, 2016 $
// $Copyright: Copyright (C) 2011-2016 Texas Instruments Incorporated -
//             http://www.ti.com/ ALL RIGHTS RESERVED $
//###########################################################################
#include "DSP28x_Project.h"     // Device Headerfile and Examples Include File

#define parity 0         // 0: no parity, 1: odd, 2: even
#define dataLength 8     // data bit length 1 - 8
#define recvBuffLen 800  // receive buffer size

// Prototype statements for functions found within this file.
__interrupt void cpu_timer0_isr(void);
__interrupt void xint3_isr(void);

// Software SCI related functions
Uint16 sw_sci_recv_byte(void);
void sw_sci_recv_bytes(Uint16* dataBuff, Uint16 len);
void sw_sci_send_byte(Uint16 data);
void sw_sci_send_bytes(Uint16* dataBuff, Uint16 len);
void sw_sci_send_string(char* txString);

// software SCI mode
// 0: rx 1: tx
Uint16 swSciMode = 0;

// rx data & state
Uint16 rxData = 0;
Uint16 rxState = 0;
Uint16 rxError = 0;

// tx data & state
Uint16 txData = 0;
Uint16 txState = 0;

void main(void)
{
// Step 1. Initialize System Control:
// PLL, WatchDog, enable Peripheral Clocks
// This example function is found in the F2806x_SysCtrl.c file.
   InitSysCtrl();

// Step 2. Initalize GPIO:
// This example function is found in the F2806x_Gpio.c file and
// illustrates how to set the GPIO to it's default state.
// InitGpio();  // Skipped for this example


// Step 3. Clear all interrupts and initialize PIE vector table:
// Disable CPU interrupts
   DINT;

// Initialize the PIE control registers to their default state.
// The default state is all PIE interrupts disabled and flags
// are cleared.
// This function is found in the F2806x_PieCtrl.c file.
   InitPieCtrl();

// Disable CPU interrupts and clear all CPU interrupt flags:
   IER = 0x0000;
   IFR = 0x0000;

// Initialize the PIE vector table with pointers to the shell Interrupt
// Service Routines (ISR).
// This will populate the entire table, even if the interrupt
// is not used in this example.  This is useful for debug purposes.
// The shell ISR routines are found in F2806x_DefaultIsr.c.
// This function is found in F2806x_PieVect.c.
   InitPieVectTable();

// Interrupts that are used in this example are re-mapped to
// ISR functions found within this file.
   EALLOW;  // This is needed to write to EALLOW protected registers
   PieVectTable.TINT0 = &cpu_timer0_isr;
   PieVectTable.XINT3 = &xint3_isr;
   EDIS;    // This is needed to disable write to EALLOW protected registers

// Step 4. Initialize the Device Peripheral. This function can be
//         found in F2806x_CpuTimers.c
   InitCpuTimers();   // For this example, only initialize the Cpu Timers
// Configure CPU-Timer 0 to interrupt every 500 milliseconds:
// 80MHz CPU Freq, 50 millisecond Period (in uSeconds)

   // Set SCI Baud Rate, 115200 bps
   ConfigCpuTimer(&CpuTimer0, 1, 781);

// To ensure precise timing, use write-only instructions to write to the entire register. Therefore, if any
// of the configuration bits are changed in ConfigCpuTimer and InitCpuTimers (in F2806x_CpuTimers.h), the
// below settings must also be updated.

   //CpuTimer0Regs.TCR.all = 0x4001; // Use write-only instruction to set TSS bit = 0
   //CpuTimer0Regs.TCR.all = 0x4000;

// Step 5. User specific code, enable interrupts:

// Configure GPIO33 as a GPIO output pin
   EALLOW;

   GpioCtrlRegs.GPBMUX1.bit.GPIO33 = 0;  // GPIO33 TX
   GpioCtrlRegs.GPBDIR.bit.GPIO33 = 1;   // output

   GpioCtrlRegs.GPBMUX1.bit.GPIO32 = 0;  // GPIO32 test pin
   GpioCtrlRegs.GPBDIR.bit.GPIO32 = 1;   // output

   GpioCtrlRegs.GPAMUX1.bit.GPIO0 = 0;  // GPIO0 RX
   GpioCtrlRegs.GPADIR.bit.GPIO0 = 0;   // input
   GpioCtrlRegs.GPAQSEL1.bit.GPIO0 = 0; // XINT1 Synch to SYSCLKOUT only

   EDIS;

   EALLOW;
   GpioIntRegs.GPIOXINT3SEL.bit.GPIOSEL = 0;   // XINT3 is GPIO0
   EDIS;

   XIntruptRegs.XINT3CR.bit.POLARITY = 0;      // Falling edge interrupt

// Enable CPU INT1 & INT12
   IER |= M_INT1;
   IER |= M_INT12;

// Enable TINT0 & XINT3
   PieCtrlRegs.PIEIER1.bit.INTx7 = 1;
   PieCtrlRegs.PIEIER12.bit.INTx1 = 1;

// Enable global Interrupts and higher priority real-time debug events:
   EINT;   // Enable Global interrupt INTM
   ERTM;   // Enable Global realtime interrupt DBGM

   GpioDataRegs.GPBSET.bit.GPIO33 = 1;   // set tx idle
   GpioDataRegs.GPBCLEAR.bit.GPIO32 = 1; // init test pin

   Uint16 tmp[recvBuffLen]; // receive buffer
   char* txString = "Software SCI EchoBack\r\n\0";

   swSciMode = 1;
   sw_sci_send_string(txString);

   // Step 6. IDLE loop. Just sit and loop forever (optional):
   while(1)
   {
	   // rx mode
	   swSciMode = 0;
	   sw_sci_recv_bytes(tmp, recvBuffLen);

	   // tx mode
	   swSciMode = 1;
	   sw_sci_send_bytes(tmp, recvBuffLen);

	   // check parity error
	   if(rxError & 0x01)
	   {
		   char* txString = "\r\nParity error\r\n\0";
		   sw_sci_send_string(txString);
	   }

	   // check stop bit error
	   if(rxError & 0x02)
	   {
		   char* txString = "\r\nStop bit error\r\n\0";
		   sw_sci_send_string(txString);
	   }
   }
}

//---------------------------------------------------------------------------
// sw_sci_send_byte:
//---------------------------------------------------------------------------
// Send one byte through software sci
// - data is the byte to be sent
// - data length is defined by dataLength
// - low dataLength bit is valid
void sw_sci_send_byte(Uint16 data)
{
	// reset tx State
	txState = 0;
	txData = data;

	// start tx timer
	CpuTimer0Regs.TCR.all = 0x4000;

	// wait until software data is sent
	// can be commented in actual use
	while(1){if(txState == 5)break;}
}

//---------------------------------------------------------------------------
// sw_sci_send_bytes:
//---------------------------------------------------------------------------
// Send n bytes through software sci
// - dataBuff is data to be sent
// - send len bytes in total
void sw_sci_send_bytes(Uint16* dataBuff, Uint16 len)
{
	Uint16 dataCnt = 0;
	while(dataCnt < len)
	{
		sw_sci_send_byte(dataBuff[dataCnt]);
		dataCnt++;
	}
}

//---------------------------------------------------------------------------
// sw_sci_send_string:
//---------------------------------------------------------------------------
// Send string through software sci
// - string must end with '\0'
// - for debug use
void sw_sci_send_string(char* txString)
{
	while(*txString != '\0')
	{
		sw_sci_send_byte(*txString);
		txString++;
	}
}

//---------------------------------------------------------------------------
// sw_sci_recv_byte:
//---------------------------------------------------------------------------
// receive one byte through software sci
// - return the byte received
// - might change rxError if error occur
Uint16 sw_sci_recv_byte(void)
{
	rxState = 0;
	XIntruptRegs.XINT3CR.bit.ENABLE = 1;

	while(1){if(rxState == 3)break;}
	return rxData;
}

//---------------------------------------------------------------------------
// sw_sci_recv_bytes:
//---------------------------------------------------------------------------
// receive len bytes through software sci
// - store result into databuff
// - might change rxError if error occur
void sw_sci_recv_bytes(Uint16* dataBuff, Uint16 len)
{
	Uint16 dataCnt = 0;
	rxError = 0; // manually set to 0

	while(dataCnt < len)
	{
		dataBuff[dataCnt] = sw_sci_recv_byte();
		dataCnt++;
	}
}

// timer0 interrupt for tx & rx state machine
__interrupt void cpu_timer0_isr(void)
{
	static Uint16 bitCount = 0; // tx & rx data bit counter
	static Uint16 par = 0;      // parity check
	static Uint16 tempData;

	// for test use
	GpioDataRegs.GPBSET.bit.GPIO32 = 1;

	// tx state machine
	if(swSciMode == 1)
	{
		// txState
		// 0: start bit, 1: data, 2: parity bit, 3: stop bit， 4： transmit done 5： idle
		switch(txState)
		{
			// tx start
			// send start bit
			case 0:
			{
				tempData = txData;
				GpioDataRegs.GPBCLEAR.bit.GPIO33 = 1; // 1 bit logic 0
				txState = 1; // move to next state
				break;
			}

			// send data
			case 1:
			{
				Uint16 bit = tempData & 0x01; // get 1 bit
				par ^= bit;                   // calculate parity

				// send 1 bit
				if(bit == 1)
					GpioDataRegs.GPBSET.bit.GPIO33 = 1;
				else
					GpioDataRegs.GPBCLEAR.bit.GPIO33 = 1;

				tempData = tempData >> 1;
				bitCount++;

				// data transmit done
				if(bitCount == dataLength)
				{
					// move to next state
					if(parity == 0)
						txState = 3;
					else
						txState = 2;
				}
				break;
			}

			// send parity bit
			case 2:
			{
				// odd
				if(parity == 1)
				{
					if(par == 0)
						GpioDataRegs.GPBSET.bit.GPIO33 = 1;
					else
						GpioDataRegs.GPBCLEAR.bit.GPIO33 = 1;
				}

				// even
				else
				{
					if(par == 1)
						GpioDataRegs.GPBSET.bit.GPIO33 = 1;
					else
						GpioDataRegs.GPBCLEAR.bit.GPIO33 = 1;
				}

				txState = 3; // move to next state
			}

			// send stop bit
			case 3:
			{
				GpioDataRegs.GPBSET.bit.GPIO33 = 1;
				txState = 4;
				break;
			}

			// transmit done
			case 4:
			{
				txState = 5;

				// reset
				bitCount = 0;
				par = 0;
				tempData = 0;
				break;
			}

			default:
				break;
		}

		CpuTimer0Regs.TCR.bit.TIF = 1; // Clear CPU-Timer Interrupt Flag
		if(txState == 5)
			CpuTimer0Regs.TCR.bit.TSS = 1; // Disable Timer 0
		PieCtrlRegs.PIEACK.all |= PIEACK_GROUP1;
	}

	// rx state machine
	else
	{
		Uint16 bit;

		// rxState
		// 0: data, 1: parity bit, 2: stop bit， 3： receive done
		switch(rxState)
		{
			// receive data
			case 0:
			{
				bit = GpioDataRegs.GPADAT.bit.GPIO0; // receive 1 bit

				tempData += bit << bitCount;
				par ^= bit;

				bitCount++;

				// data receive done
				if(bitCount == dataLength)
				{
					// move to next state
					if(parity == 0)
						rxState = 2;
					else
						rxState = 1;
				}
				break;
			}

			// receive parity bit
			case 1:
			{
				bit = GpioDataRegs.GPADAT.bit.GPIO0;
				par ^= bit;

				// check parity error
				if((parity == 1 && par == 0) || (parity == 2 && par == 1))
					rxError |= 0x01;

				rxState = 2;
				break;
			}

			// check stop bit
			case 2:
			{
				bit = GpioDataRegs.GPADAT.bit.GPIO0;
				if(bit != 1)
					rxError |= 0x02;

				// reset
				rxState = 3;
				bitCount = 0;
				par = 0;
				rxData = tempData;
				tempData = 0;

				break;
			}

			default:
				break;
		}

		CpuTimer0Regs.TCR.bit.TIF = 1; // Clear CPU-Timer Interrupt Flag
		if(rxState == 3)
			CpuTimer0Regs.TCR.bit.TSS = 1; // Disable Timer 1
		// Acknowledge this interrupt
		PieCtrlRegs.PIEACK.all |= PIEACK_GROUP1;
	}

	// for test use
	GpioDataRegs.GPBCLEAR.bit.GPIO32 = 1;
}

// detect stop bit for rx
__interrupt void xint3_isr(void)
{
	if(rxState == 0)
	{
		DELAY_US(4);
		CpuTimer0Regs.TCR.all = 0x4000; // Start timer0
		XIntruptRegs.XINT3CR.bit.ENABLE = 0; // Disable XINT3
	}

	// Acknowledge this interrupt
	PieCtrlRegs.PIEACK.all |= PIEACK_GROUP12;
}
//===========================================================================
// No more.
//===========================================================================
